#!/usr/bin/env ruby

require 'csv'

Database = Struct.new(:attrs, :data)

DTreeNode = Struct.new(:attr_idx, :edges)
DTEdge = Struct.new("DTEdge", :dest, :value)
DTreeLeaf = String

def incr_hash(hash, key)
  hash[key] = 0 if hash[key].nil? ; hash[key] += 1
end

def split_by_attr(db, samples, attr)
  groups = {}
  samples.each { |s|
    groups[db.data[s][attr]] = [] if groups[db.data[s][attr]].nil?
    groups[db.data[s][attr]] << s
  }
  return groups
end

def load_csv(file)
  reader = CSV::Reader.parse(File.open(file, 'rb'))
  attrs = reader.shift
  data = [] ; reader.each { |row| data << row }
  return Database.new(attrs, data)
end

# Is there only one class in db?
def one_class?(db, samples)
  samples.each { |s| 
    return false if db.data[s][-1] != db.data[samples[0]][-1] 
  }
  return true
end

# Entropy of a set of samples
def entropy (db, samples)
  classes = {} ; samples.each { |s| incr_hash(classes, db.data[s][-1]) }
  n = samples.length.to_f
  entropy = 0.0 ; classes.each_value { |v| entropy -= (v/n) * Math.log(v/n) }
  return entropy
end

# Entropy gain by partitioning using attr
def entropy_gain(db, samples, attr)
  attr_groups = split_by_attr(db, samples, attr)
  new_entropy = 0.0
  attr_groups.each_value { |g|
    new_entropy += (g.length.to_f / samples.length) * entropy(db, g)
  }
  return [attr_groups, entropy(db, samples) - new_entropy]
end

# attrs: list of attr indexes still available
# samples: list of db element indexes
def build_simple_tree_impl (db, attrs, samples)
  if one_class?(db, samples) || (attrs.length == 0)
    return DTreeLeaf.new(db.data[samples[0]][-1])
  end
  attr_idx = attrs[0]
  attrs = attrs.clone ; attrs.delete_at(0)
  node = DTreeNode.new(attr_idx, [])
  groups = split_by_attr(db, samples, attr_idx)
  groups.each { |attr_val, elts|
    node.edges << DTEdge.new(build_simple_tree_impl(db, attrs, elts), attr_val)
  }
  return node
end

# Create a decision tree naively.
def build_simple_tree(db)
  build_simple_tree_impl(db,
                         (0...db.attrs.length-1).to_a, 
                         (0...db.data.length).to_a)
end

# attrs: list of attr indexes still available
# samples: list of db element indexes
def id3_impl (db, attrs, samples)
  if one_class?(db, samples) || (attrs.length == 0)
    return DTreeLeaf.new(db.data[samples[0]][-1])
  end
  best_gain = -1.0 ; attr_idx = 0; groups = {}
  attrs.each { |attr| 
    res = entropy_gain(db, samples, attr)
    if res[1] > best_gain
      groups = res[0] ; attr_idx = attr ; best_gain = res[1]
    end
  }
  attrs = attrs.clone ; attrs.delete_at(attr_idx)
  node = DTreeNode.new(attr_idx, [])
  groups.each { |attr_val, elts|
    node.edges << DTEdge.new(build_simple_tree_impl(db, attrs, elts), attr_val)
  }
  return node
end

# Create a decision tree using id3.
def id3(db) 
  id3_impl(db, (0...db.attrs.length-1).to_a, (0...db.data.length).to_a)
end

def name_of(db, node)
  return db.attrs[node.attr_idx] if node.kind_of?(DTreeNode)
  return node
end

def print_tree(db, tree)
  f = File.new('/tmp/output.dot', 'w')
  f << "graph decision_tree {\n"
  def print_node(db, f, node, orig)
    return orig + 1 if !node.kind_of?(DTreeNode)
    dest = orig + 1
    node.edges.each { |edge|
      f << "node#{dest} [label=\"#{name_of(db, edge.dest)}"
      f << '?' if edge.dest.kind_of?(DTreeNode)
      f << '"];' << "\n"
      f << "node#{orig} -- node#{dest} [label=\"#{edge.value}\"]\n"
      dest = print_node(db, f, edge.dest, dest)
    }
    return dest
  end
  f << 'node0 [label = "' << name_of(db, tree) << "?\"];\n"
  print_node(db, f, tree, 0)
  f << "}\n"
  f.close
  `dot -Tps /tmp/output.dot | gv -`
end

def get_class(db, tree, sample)
  return tree if tree.kind_of?(DTreeLeaf)
  tree.edges.each { |edge|
    if db.data[sample][tree.attr_idx] == edge.value
      return get_class(db, edge.dest, sample)
    end
  }
  return "Unknown"
end

def test_quality(db, tree)
  nb_good = 0
  for sample in 0...db.data.length
    res = get_class(db, tree, sample)
    p "#{db.data[sample][-1]} => #{res}"
    nb_good += 1 if res == db.data[sample][-1]
  end
  return nb_good.to_f / db.data.length
end

if (ARGV.length != 1)
  STDERR.puts "Usage: id3.rb data.csv\n"
  exit 1
end

db = load_csv(ARGV[0])

# t = build_simple_tree(db)
# print_tree(db, t)

id_tree = id3(db)
print_tree(db, id_tree)
